/**
 * @file popup
 * @author songjunke & luowenwen02
 * @since 2018-08-21
 */

/* eslint-disable no-use-before-define */

/* global chrome */
// 定义常量
const UA_MOBILE = ''
    + 'Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X) '
    + 'AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0 Mobile/15A372 Safari/604.1';

const REG_HM_JS = /\/\/hm\.baidu\.com\/(hm?\.js)\?(\w+)/i;
const REG_HM_GIF = /\/\/hm\.baidu\.com\/hm\.gif/gi;

let heatmapUserAgent = 'pc';

let uaType = 'pc';
let metaData = {};
let filterMetaData = {};
let utils = {
    isObject: obj => {
        let type = typeof obj;

        return type === 'function' || type === 'object' && !!obj;
    },
    isString: obj => Object.prototype.toString.call(obj) === '[object String]',
    isEmptyObject: obj => Object.keys(obj).length === 0,
    // 判断是否为 json 字符串
    isJsonString(str) {
        if (!utils.isString(str)) {
            return false;
        }

        try {
            JSON.parse(str);

            return true;
        }
        catch (e) {
            return false;
        }
    },
    // queryString转对象
    parseUrl(url) {
        let queryString = url.split('?')[1];
        let ret = {};
        let reg = /([^&]*)=([^&]*)/g;
        let matches;

        while ((matches = reg.exec(queryString))) {
            let key = matches[1];
            let val = matches[2];

            // 如果是多值
            if (ret[key]) {
                ret[key] = Array.isArray(ret[key]) ? ret[key] : [ret[key]];
                ret[key].push(
                    utils.isJsonString(decodeURIComponent(val))
                        ? JSON.parse(decodeURIComponent(val))
                        : decodeURIComponent(val)
                );
            }
            else {
                ret[key] = utils.isJsonString(decodeURIComponent(val))
                    ? JSON.parse(decodeURIComponent(val))
                    : decodeURIComponent(val);
            }
        }

        return ret;
    },
    // 转化responseHeaders格式
    formatData(data) {
        let obj = {};

        data.forEach(function (item) {
            obj[item.name] = item.value;
        });

        return obj;
    }
};

// 如果发生路由变化，可能有三种情况,js跳转，302跳转，导航栏切换url，重置
chrome.webNavigation.onBeforeNavigate.addListener(function (e) {
    if (e.frameId === 0) {
        let tabId = e.tabId;

        // 重置当前 tab 中存数据
        metaData[tabId] = {
            data: {}
        };

        // 重置当前 tab 中存数据
        filterMetaData[tabId] = {
            data: {},
            result: null
        };
        chrome.runtime.sendMessage({
            type: 'RESET'
        });
    }

});

chrome.webNavigation.onCompleted.addListener(function (e) {
    chrome.runtime.sendMessage({
        type: 'REFRESH'
    });
});
// 当标签页关闭时清空缓存
chrome.tabs.onRemoved.addListener(function (tabId) {
    if (metaData[tabId]) {
        metaData[tabId] = 'empty';
    }
});

// 发送requestHeaders
chrome.webRequest.onBeforeSendHeaders.addListener(function (e) {

    if (e.frameId !== 0) {
        if (heatmapUserAgent === 'mobile') {
            for (let i = 0, l = e.requestHeaders.length; i < l; i++) {
                if (e.requestHeaders[i].name === 'User-Agent') {
                    e.requestHeaders[i].value = UA_MOBILE;
                    break;
                }
            }

            return {
                requestHeaders: e.requestHeaders
            };
        }

        return;
    }

    let url = e.url;
    let tabId = e.tabId;
    // 用于同步 requestHeaders 和 reponseHeaders
    let requestId = e.requestId;
    let referer = '';
    let si;

    // 请求中存在 si 的情况：
    // 1. 既有 js , 又有 gif ；2. 只有 js ；3. 只有 gif ;
    if (url.match(REG_HM_JS)) {
        let regInfo = url.match(REG_HM_JS);
        let jsType = regInfo[1];
        si = regInfo[2];

        let requestHeaders = utils.formatData(e.requestHeaders || []);
        // hm.js 或 h.js 返回内容为空情况很多，此处先判断下是否不存在 referer 的情况
        referer = requestHeaders.Referer || '';

        if (!metaData[tabId]) {
            metaData[tabId] = {
                data: {}
            };
        }

        // 把si作为key存在tabId中
        if (!metaData[tabId].data[si]) {
            metaData[tabId].data[si] = {
                j: [],
                g: []
            };
        }

        let queryObj = {
            type: jsType,
            url: url,
            referer: referer,
            requestId: requestId,
            si: si
        };

        metaData[tabId].data[si].j.push(queryObj);
    }
    else if (url.match(REG_HM_GIF)) {
        let requestHeaders = utils.formatData(e.requestHeaders || []);
        referer = requestHeaders.Referer || '';

        let queryObj = utils.parseUrl(url);
        si = queryObj.si;

        // 在 tab 关闭的时候会发送 hm.gif 请求，此时把该请求过滤
        if (metaData[tabId] === 'empty') {
            delete metaData[tabId];
            return;
        }

        if (!metaData[tabId]) {
            metaData[tabId] = {
                data: {}
            };
        }

        // 把si作为key存在tabId中
        if (!metaData[tabId].data[si]) {
            metaData[tabId].data[si] = {
                j: [],
                g: []
            };
        }

        queryObj.url = url;
        queryObj.referer = referer;
        queryObj.requestId = requestId;

        metaData[tabId].data[si].g.push(queryObj);
    }

    if (uaType === 'mobile') {
        // 修改 ua 为 mobile
        for (let i = 0, l = e.requestHeaders.length; i < l; i++) {
            if (e.requestHeaders[i].name === 'User-Agent') {
                e.requestHeaders[i].value = UA_MOBILE;
                break;
            }
        }

        return {
            requestHeaders: e.requestHeaders
        };
    }

}, {
    urls: ['<all_urls>']
}, ['blocking', 'requestHeaders']);

let initMetaObject = (tabId, si) => {
    if (!filterMetaData[tabId]) {
        filterMetaData[tabId] = {
            data: {}
        };
    }

    // 接受到新请求，清空缓存结果
    // filterMetaData[tabId].result = null;

    // 把si作为key存在tabId中
    if (!filterMetaData[tabId].data[si]) {
        filterMetaData[tabId].data[si] = {
            j: [],
            g: []
        };
    }
};
// 获取reposeHeaders
chrome.webRequest.onHeadersReceived.addListener(function (e) {
    let url = e.url;
    let tabId = e.tabId;
    let requestId = e.requestId;

    if (e.frameId !== 0) {
        // 判断 iframe 是否禁止被加载
        if (/jn=pageclick/.test(e.url)) {
            let responseHeaders = utils.formatData(e.responseHeaders || []);
            let isIFrameForbidden = false;

            if (responseHeaders['X-Frame-Options']
                && responseHeaders['X-Frame-Options'].toLowerCase() !== 'allowall') {
                metaData[tabId].isIFrameForbidden = true;
                filterMetaData[tabId].isIFrameForbidden = true;
                isIFrameForbidden = true;
            }

            chrome.tabs.query({active: true, currentWindow: true}, function(tabs) {
                chrome.tabs.sendMessage(tabId, {
                    type: 'IFRAME',
                    data: {
                        isIFrameForbidden: isIFrameForbidden
                    }
                });
            });

        }

        return;
    }

    if ((!url.match(REG_HM_JS)
        && !url.match(REG_HM_GIF))
        || (!(+e.statusCode === 200 || +e.statusCode === 304))
        || e.type === 'ping'
    ) {
        return;
    }
    // 是否存在 js 脚本
    if (url.match(REG_HM_JS)) {
        let responseHeaders = utils.formatData(e.responseHeaders || []);
        let regInfo = url.match(REG_HM_JS);
        let si = regInfo[2];
        // 判断 Content-Length 的长度
        let contentLength = responseHeaders['Content-Length'];
        if (e.statusCode === 304) {
            // 304默认给一个
            contentLength = 1;
        }
        metaData[tabId] && metaData[tabId].data[si] && metaData[tabId].data[si].j.forEach(function (item, index) {
            if (item.requestId === requestId) {
                metaData[tabId].data[si].j[index].contentLength = contentLength;

                initMetaObject(tabId, si);

                filterMetaData[tabId].data[si].j.push(metaData[tabId].data[si].j[index]);
                sendRequestMessage(Object.assign(item, {type: 'j'}), tabId);
            }
        });

        return;
    }

    if (url.match(REG_HM_GIF)) {
        let queryObj = utils.parseUrl(url);
        let si = queryObj.si;

        metaData[tabId] && metaData[tabId].data[si] && metaData[tabId].data[si].g.forEach(function (item) {
            if (item.requestId === requestId) {
                initMetaObject(tabId, si);

                filterMetaData[tabId].data[si].g.push(item);
                sendRequestMessage(Object.assign(item, {type: 'g'}), tabId);
            }

        });
    }

}, {
    urls: ['<all_urls>']
}, ['blocking', 'responseHeaders']);

chrome.runtime.onMessage.addListener(function (message, sender, sendResponse) {
    if (message.type === 'SETUA') {
        // 监听来自 content_script.js 的消息
        setUaType(message.uaType);
        sendResponse('SETUADONE');
    }

    if (message.type === 'HEATMAP_SETUA') {
        // 监听来自 content_heatmap.js 的消息
        heatmapUserAgent = message.uaType;
        sendResponse('HEATMAP_SETUA_DONE');
    }

});

function sendRequestMessage(data, tabId) {
    chrome.windows.getCurrent(function (window) {
        let wid = window.id;

        chrome.tabs.query({active: true, windowId: wid}, function (tab) {
            if (tabId === tab[0].id) {
                chrome.runtime.sendMessage({
                    type: 'RECEIVEDREQUEST',
                    data: JSON.stringify(data)
                });
            }
        });
    });
}
// 返回根据tab信息返回元数据
function getMetaDataByTab(tab) {
    // 过滤 metaData 中空对象
    return metaData[tab.id];
}
// 返回元数据
function getMetaData() {
    // 过滤 metaData 中空对象
    let data = {};
    for (let i in filterMetaData) {
        if (filterMetaData.hasOwnProperty(i)) {
            if (filterMetaData[i].data && !utils.isEmptyObject(filterMetaData[i].data)) {
                data[i] = filterMetaData[i];
            }
        }

    }

    return data;
}

function getCheckResultBySiList(params, callback) {
    let siList = params.siList;

    $.ajax({
        method: 'post',
        url: 'https://tongji.baidu.com/sc-web/ajax/post',
        dataType: 'json',
        data: {
            si: siList,
            method: 'home/js/check/signatureCheck'
        },
        headers: {
            'X-Requested-With': 'XMLHttpRequest'
        }
    })
        .done(function (data, textStatus, jqXHR) {
            try {
                callback(data, data.status);
            }
            catch (e) {
                console.log('error');
            }
        })
        .fail(function (jqXHR, textStatus, errorThrown) {
            console.log('error');
        });
}
function cacheResult(tabId, result) {
    filterMetaData[tabId].result = result;
}
function resetCacheResult(tabId) {
    filterMetaData[tabId].result = null;
}

function setUaType(type) {
    uaType = type;
}

// 返回 uaType
function getUaType() {
    return uaType || 'pc';
}
